/***************************************************************************
 * Copyright 2012 by
 *  + Christian-Albrechts-University of Kiel
 *    + Department of Computer Science
 *      + Software Engineering Group 
 *  and others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.common;

import java.io.IOException;
import java.net.MalformedURLException;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;

import kieker.analysis.AnalysisController.STATE;
import kieker.analysis.model.analysisMetaModel.MIDependency;
import kieker.analysis.model.analysisMetaModel.MIProject;
import kieker.analysis.plugin.AbstractPlugin;
import kieker.analysis.repository.AbstractRepository;
import kieker.webgui.common.exception.AnalysisStateException;
import kieker.webgui.common.exception.DisplayNotFoundException;
import kieker.webgui.common.exception.LibraryAlreadyExistingException;
import kieker.webgui.common.exception.LibraryLoadException;
import kieker.webgui.common.exception.NewerProjectException;
import kieker.webgui.common.exception.ProjectAlreadyExistingException;
import kieker.webgui.common.exception.ProjectNotExistingException;
import kieker.webgui.common.util.ACManager;
import kieker.webgui.common.util.FSManager;
import kieker.webgui.common.util.PluginFinder;

import org.primefaces.model.UploadedFile;

/**
 * This is an implemented facade for the project management. For technical reasons it is a singleton class.
 * 
 * @author Nils Christian Ehmke
 * @version 1.0
 */
public final class ProjectManagerFacade implements IProjectManagerFacade {

	private static final ProjectManagerFacade INSTANCE = new ProjectManagerFacade();
	private final ConcurrentHashMap<String, Object> fileSystemLocks = new ConcurrentHashMap<String, Object>();
	private final ConcurrentHashMap<String, Object> analysesLocks = new ConcurrentHashMap<String, Object>();

	/**
	 * This is the private constructor for this class, as this class is a singleton.
	 */
	private ProjectManagerFacade() {
		// No code necessary.
	}

	/**
	 * This method delivers the one and only singleton instance of this class.
	 * 
	 * @return The singleton instance.
	 */
	public static ProjectManagerFacade getInstance() {
		return ProjectManagerFacade.INSTANCE;
	}

	@Override
	public void addProject(final String projectName) throws ProjectAlreadyExistingException, IOException {
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);

		synchronized (projectLock) {
			FSManager.getInstance().addProject(projectName);
		}
	}

	@Override
	public void copyProject(final String originalProjectName, final String newProjectName) throws ProjectNotExistingException, ProjectAlreadyExistingException,
			IOException {
		// Get the file system locks for both projects
		final Object srcProjectLock = this.getLock(originalProjectName, this.fileSystemLocks);
		final Object dstProjectLock = this.getLock(newProjectName, this.fileSystemLocks);

		// We cannot risk just using two synchronized-blocks as such a thing could block the application. We therefore have to sort the locks by name in order to
		// make sure that two methods would use the same order to access (if project A has to be renamed to B and B has to be renamed to A, we would always
		// synchronize on A before synchronizing on B).
		final Object lockFst;
		final Object lockSnd;
		if (originalProjectName.compareTo(newProjectName) < 0) {
			lockFst = srcProjectLock;
			lockSnd = dstProjectLock;
		} else {
			lockFst = dstProjectLock;
			lockSnd = srcProjectLock;
		}

		synchronized (lockFst) {
			synchronized (lockSnd) {
				FSManager.getInstance().copyProject(originalProjectName, newProjectName);
			}
		}
	}

	@Override
	public MIProject openProject(final String projectName) throws ProjectNotExistingException, IOException {
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);

		synchronized (projectLock) {
			return FSManager.getInstance().openProject(projectName);
		}
	}

	@Override
	public Object openProject(final String projectName, final ClassAndMethodContainer classAndMethodContainer) throws ProjectNotExistingException, IOException {
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);

		synchronized (projectLock) {
			return FSManager.getInstance().openProject(projectName, classAndMethodContainer);
		}
	}

	@Override
	public void saveProject(final String projectName, final MIProject project, final long timeStamp, final boolean overwriteNewerProject) throws
			ProjectNotExistingException, IOException, NewerProjectException {
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);

		synchronized (projectLock) {
			FSManager.getInstance().saveProject(projectName, project, timeStamp, overwriteNewerProject);
		}
	}

	@Override
	public long getCurrTimeStamp(final String projectName) throws ProjectNotExistingException {
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);

		synchronized (projectLock) {
			return FSManager.getInstance().getCurrTimeStamp(projectName);
		}
	}

	@Override
	public void uploadLibrary(final UploadedFile file, final String projectName) throws ProjectNotExistingException, IOException, LibraryAlreadyExistingException {
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);

		synchronized (projectLock) {
			FSManager.getInstance().uploadLibrary(file, projectName);
		}
	}

	@Override
	public ClassLoader getClassLoader(final String projectName) throws ProjectNotExistingException, IOException {
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);

		synchronized (projectLock) {
			return FSManager.getInstance().getClassLoader(projectName);
		}
	}

	@Override
	public List<Class<AbstractRepository>> getAllRepositoriesWithinLib(final MIDependency lib, final String projectName, final ClassLoader classLoader,
			final ClassAndMethodContainer classAndMethodContainer) throws LibraryLoadException {
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);

		synchronized (projectLock) {
			try {
				return PluginFinder.getAllRepositoriesWithinJar(FSManager.getInstance().getURL(lib, projectName), classLoader, classAndMethodContainer);
			} catch (final MalformedURLException ex) {
				throw new LibraryLoadException("An error occured while loading the library.", ex);
			}
		}
	}

	@Override
	public List<Class<AbstractPlugin>> getAllPluginsWithinLib(final MIDependency lib, final String projectName, final ClassLoader classLoader,
			final ClassAndMethodContainer classAndMethodContainer) throws LibraryLoadException {
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);

		synchronized (projectLock) {
			try {
				return PluginFinder.getAllPluginsWithinJar(FSManager.getInstance().getURL(lib, projectName), classLoader, classAndMethodContainer);
			} catch (final MalformedURLException ex) {
				throw new LibraryLoadException("An error occured while loading the library.", ex);
			}
		}
	}

	@Override
	public List<Class<AbstractRepository>> getAllRepositoriesWithinKiekerLib(final ClassLoader classLoader, final ClassAndMethodContainer classAndMethodContainer)
			throws LibraryLoadException {
		try {
			return PluginFinder.getAllRepositoriesWithinJar(FSManager.getInstance().getKiekerURL(), classLoader, classAndMethodContainer);

		} catch (final NullPointerException ex) {
			throw new LibraryLoadException("An error occured while loading the library.", ex);
		}
	}

	@Override
	public List<Class<AbstractPlugin>> getAllPluginsWithinKiekerLib(final ClassLoader classLoader, final ClassAndMethodContainer classAndMethodContainer)
			throws LibraryLoadException {
		try {
			return PluginFinder.getAllPluginsWithinJar(FSManager.getInstance().getKiekerURL(), classLoader, classAndMethodContainer);

		} catch (final NullPointerException ex) {
			throw new LibraryLoadException("An error occured while loading the library.", ex);
		}
	}

	@Override
	public List<String> listAllLibraries(final String projectName) throws ProjectNotExistingException {
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);

		synchronized (projectLock) {
			return FSManager.getInstance().listAllLibraries(projectName);
		}
	}

	@Override
	public Collection<String> listAllProjects() {
		return FSManager.getInstance().listAllProjects();
	}

	@Override
	public void initializeAnalysis(final String projectName, final ClassLoader classLoader) throws ProjectNotExistingException, AnalysisStateException {
		// We have to lock both - the project and the analysis, as a file has to be loaded
		final Object projectLock = this.getLock(projectName, this.fileSystemLocks);
		final Object analysisLock = this.getLock(projectName, this.analysesLocks);

		synchronized (projectLock) {
			synchronized (analysisLock) {
				ACManager.getInstance().initializeAnalysis(projectName, classLoader);
			}
		}
	}

	@Override
	public void cleanAnalysis(final String projectName) throws ProjectNotExistingException, AnalysisStateException {
		final Object analysisLock = this.getLock(projectName, this.analysesLocks);

		synchronized (analysisLock) {
			ACManager.getInstance().cleanAnalysis(projectName);
		}
	}

	@Override
	public void startAnalysis(final String projectName) throws ProjectNotExistingException, AnalysisStateException {
		final Object analysisLock = this.getLock(projectName, this.analysesLocks);

		synchronized (analysisLock) {
			ACManager.getInstance().startAnalysis(projectName);
		}
	}

	@Override
	public void stopAnalysis(final String projectName) throws ProjectNotExistingException, AnalysisStateException {
		final Object analysisLock = this.getLock(projectName, this.analysesLocks);

		synchronized (analysisLock) {
			ACManager.getInstance().stopAnalysis(projectName);
		}
	}

	@Override
	public Object getDisplay(final String projectName, final String viewName, final String displayName) throws ProjectNotExistingException, DisplayNotFoundException {
		final Object analysisLock = this.getLock(projectName, this.analysesLocks);

		synchronized (analysisLock) {
			return ACManager.getInstance().getDisplay(projectName, viewName, displayName);
		}
	}

	@Override
	public STATE getCurrentState(final String projectName) throws ProjectNotExistingException {
		final Object analysisLock = this.getLock(projectName, this.analysesLocks);

		synchronized (analysisLock) {
			return ACManager.getInstance().getCurrentState(projectName);
		}
	}

	/**
	 * This method can be used to get the lock for a given project. If the lock doesn't exist already, a new one will be created.
	 * 
	 * @param projectName
	 *            The project whose lock should be delivered.
	 * @param lockMap
	 *            The map containing the locks.
	 * @return A lock object for the project.
	 */
	private Object getLock(final String projectName, final ConcurrentHashMap<String, Object> lockMap) {
		// Create a new lock and put it into the map - if it doesn't already contain the key
		final Object newLock = new Object();
		final Object existLock = lockMap.putIfAbsent(projectName, newLock);

		// Find out which of the both locks is the correct one
		final Object lock;
		if (existLock != null) {
			lock = existLock;
		} else {
			lock = newLock;
		}

		return lock;
	}
}
